import coloredlogs, logging
from DecenTT.env.config import LOG_PATH

logging.basicConfig(filename= f"{LOG_PATH}",format='%(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)
logger.setLevel(level=logging.DEBUG)
coloredlogs.install(level="DEBUG", logger=logger)



class BootstrapNodeError(Exception):
    """
        Subjected: Bootstrap Node reading failure
    """
    def __init__(self, error:Exception = Exception("Unidentified Internal Bootstrap Error")):
        self.error = error

    def __str__(self):
        return f"Bootstrap Identification Failure : {self.error}"

class InvalidBootstrap(Exception):
    """
        Subjected: Bootstrap node specified is invalid
    """


def load_env(path: str) -> None:
    from dotenv import load_dotenv
    from tqdm import tqdm
    from DecenTT.env.logger import logger

    try:
        try:
            logger.info("Loading Environment Variables")
            for i in tqdm(range(100)):
                pass
        except Exception as e:
            logger.debug("Error in TQDM, kindly check the package")
        load_dotenv(path)
        logger.info("Environment Variables Loaded")

    except Exception as e:
        logger.error("Error in loading Environment Variables")
        logger.error(str(e))



import threading
import multiprocessing


class ThreadDecorator:
    @classmethod
    def thread(cls, function):  #
        def inner_thread(*args):
            fun_thread = threading.Thread(target=function, args=args)
            fun_thread.start()
            return fun_thread

        return inner_thread

class ProcessDecorator:
    @classmethod
    def thread(cls, function):
        def inner_thread(*args):
            fun_thread = multiprocessing.Process(target=function, args=args)
            fun_thread.start()
            return fun_thread

        return inner_thread



from base64 import b64decode

class IPFSResponse:
    """
        Sample JSON Response from IPFS HTTP API:
            {
                "from":"ACQIARIgvsDLWj3HNX2jJQGZ5H1Ij1RJfOXDj/tshfztcgwO57M=",
                "data":"aGVsbG8gd29ybGRz",
                "seqno":"Frs0H9ZNr5Y=",
                "topicIDs":["abc"]
            }
    """
    class UserData:
        def __init__(self, from_user:str, seq_no:str)-> None:
            self.from_user = from_user
            self.seq_no=seq_no

    class Message:
        class Payload:
            def __init__(self, data:str):
                self.__data = data

            def decode(self)->str:
                return b64decode(self.__data).decode("utf-8")

            def __str__(self):
                return self.__data


        def __init__(self, topic:str, data:str)->None:
            self.topic = topic
            self.payload = self.Payload(data=data)
            self.qos = None
            self.retain = None

    def __init__(self, in_dict:dict) -> None:
        self.raw_dict = in_dict
        self.userdata = self.UserData(from_user=self.raw_dict["from"], seq_no=self.raw_dict["seqno"])
        self.msg = self.Message(topic=self.raw_dict["topicIDs"][0], data=self.raw_dict["data"])




"""
Module:             IPFS Client
SubModule:          Config
Author:             Amey Mahadik <https://github.com/saapo-ka-baadshah>
"""
import wget
import os
import platform
from DecenTT.IPFS.IPFSExceptions import UnsupportedPlatform
from DecenTT.IPFS.logger import logger
import zipfile
import tarfile
import sys
import subprocess

import shutil

# The endpoints go here:
ENDPOINTS = {
    "pubsub": {
        "sub": "/api/v0/pubsub/sub",
        "pub": "/api/v0/pubsub/pub"
    }
}

# The Setup Class goes here:
class Setup:
    def __init__(self) -> None:
        self.cwd = os.getcwd()

    @staticmethod
    def __get_platform() -> str:
        return str(platform.system())

    def __get_os_name(self) -> str:
        pf = self.__get_platform().lower()

        if "windows" in pf:
            return "windows"
        elif "linux" in pf:
            return "linux"
        elif "darwin" in pf:
            return "macos"
        else:
            raise UnsupportedPlatform

    @staticmethod
    def __bar_progress(current, total, width=80):
        progress_message = "Downloading: %d%% [%d / %d] bytes" % (current / total * 100, current, total)
        # Don't use print() as it will print in new line every time.
        sys.stdout.write("\r" + progress_message)
        sys.stdout.flush()

    @staticmethod
    def __run_ps(cmd):
        completed = subprocess.run(["powershell", "-Command", cmd], capture_output=True)
        return completed

    def __windows_installer(self):
        temp_path = os.path.join(self.cwd, "temp")
        os.mkdir(temp_path)
        fileName = wget.download(
            url="https://dist.ipfs.io/go-ipfs/v0.10.0/go-ipfs_v0.10.0_windows-amd64.zip",
            out=str(temp_path),
            bar=self.__bar_progress
        )
        path_to_zip_file = os.path.join(temp_path, fileName)

        with zipfile.ZipFile(path_to_zip_file, 'r') as zip_ref:
            zip_ref.extractall(temp_path)

        package_dir = os.path.expanduser(os.path.join(temp_path, "go-ipfs"))
        documents_path = os.path.expanduser("~/Documents")
        profile_path = os.path.join(documents_path, "WindowsPowerShell", "profile.ps1")
        command = f'Add-Content {profile_path} "[System.Environment]::SetEnvironmentVariable(`PATH`,`$Env:PATH+`;;' + \
                  f'{package_dir}`)" '
        self.__run_ps(cmd=command)
        logger.log(f"Installed: IPFS")

    def __linux_installer(self):
        temp_path = os.path.join(self.cwd, "temp")
        cur_file_path = os.path.dirname(os.path.realpath(__file__))
        os.mkdir(temp_path)
        fileName = wget.download(
            url="https://dist.ipfs.io/go-ipfs/v0.10.0/go-ipfs_v0.10.0_linux-amd64.tar.gz",
            out=str(temp_path),
            bar=self.__bar_progress
        )

        path_to_tar_file = os.path.join(temp_path, fileName)

        tar = tarfile.open(path_to_tar_file)
        tar.extractall(path=temp_path)
        logger.info("Tarfile Extracted")
        subprocess.call(["sudo", "bash", f"{temp_path}/go-ipfs/install.sh"])
        subprocess.call(["sudo", "ipfs", "init"])
        logger.info(f"Installed: IPFS")
        # subprocess.call(["sudo", "mkdir", f"/var/log/ipfs"])
        # subprocess.call(["sudo", "touch", f"/var/log/ipfs/ipfs.log"])
        # subprocess.call(["sudo", "touch", f"/var/log/ipfs/error.log"])
        # subprocess.call(["sudo", "cp", f"{cur_file_path}/ipfs.service", "/etc/systemd/system"])
        # logger.info(f"Installed: IPFS as Service")

        logger.info(f"Installation: Completed !")

    def install(self, os_name: str = None) -> None:
        temp_path = os.path.join(self.cwd, "temp")
        if os.path.exists(temp_path) and os.path.isdir(temp_path):
            shutil.rmtree(temp_path)
        else:
            pass

        if not os_name:
            os_name = self.__get_os_name()
        else:
            os_name = str(os_name).lower()
        if os_name == "windows":
            self.__windows_installer()
        elif os_name == "macos":
            pass
        elif os_name == "linux":
            try:
                logger.info("Calling linux installer")
                self.__linux_installer()
            except Exception as e:
                logger.error(f"Error:{e}")
                temp_dir = os.path.join(self.cwd, "temp")
                if os.path.exists(temp_dir) and os.path.isdir(temp_dir):
                    shutil.rmtree(temp_dir)

        else:
            raise UnsupportedPlatform



import os.path

from DecenTT.IPFS.locales.env_config import getBootstrapNodes
import requests
from DecenTT.IPFS.locales.IPFSLocalesExceptions import *
from DecenTT.IPFS.logger import logger


class BStrapper:
    """
    >>> bstrapper = BStrapper()
    >>> # bstrapper.add("/dnsaddr/bootstrap.libp2p.io/p2p/QmcZf59bWwK5XFi76CZX8cbJ4BhTzzA3gU1ZjYZcYW3dwt")
    >>> bstrapper.load_all()
    """

    def __init__(self, path: str = None) -> None:
        self.__env_path = path

    def add(self, address: str):
        valid = self.__validate(address)
        if valid:
            resp = requests.post(
                f"http://127.0.0.1:5001/api/v0/bootstrap/add?arg={address}"
            )
            logger.info(f"Bootstrap Addition Response: {resp}")
            logger.info(f"Bootstrap Added : {address}")
        else:
            raise InvalidBootstrap

    def get_peerId(self, address: str):
        domain_list = address.split("/")
        return domain_list[-1]

    def record(self, address: str):
        valid = self.__validate(address=address)
        if valid:
            bs_list = getBootstrapNodes(self.__env_path)
            last_index = len(bs_list)
            file_path = os.path.dirname(os.path.realpath(__file__))
            with open(f"{file_path}/.env", 'ab') as f:
                f.write(f"BOOTSTRAP_{last_index}={address}\n")

    def load_all(self):
        bs_list = getBootstrapNodes(self.__env_path)
        for bs in bs_list:
            self.add(address=bs)

    def __validate(self, address: str) -> bool:
        peerId = self.get_peerId(address=address)
        seg_len = len(address.split("/"))

        if peerId == "" or seg_len < 2:
            return False
        return True



from DecenTT.MQTT.client import Client
from DecenTT.MQTT.logger import logging, logger
from DecenTT.env.config import SUBSYSTEM_PATH


class Subscriber:
    def __init__(self, client:Client, topic:any, callback:any = None, qos:int = 1, log_path:str = f"{SUBSYSTEM_PATH}/subscriber.log", log_level:str="DEBUG") -> None:
        self.__client = client.client
        if(not(callback == None)):
            self.__client.on_message = callback
        else:
            self.__client.on_message = self.__callback
        self.__client.subscribe(topic=topic, qos=qos)

        self.__log_path = log_path
        self.__log_level = log_level
        self.__file_handler = logging.FileHandler(self.__log_path)
        levels = {
            "DEBUG": logging.DEBUG,
            "INFO": logging.INFO,
            "ERROR": logging.ERROR,
            "WARNING": logging.WARNING,
            "CRITICAL": logging.CRITICAL,
        }
        self.__file_handler.setLevel(level=levels[self.__log_level])
        logger.addHandler(self.__file_handler)

    def __callback(self, client, userdata, msg):
        logger.debug(f"From {client}:\t {msg.topic} got :\t {(msg.payload)}")

    def stop_logging(self):
        logger.removeHandler(self.__file_handler)




import paho.mqtt.client as mqtt
from DecenTT.env.config import MQTT_HOST, MQTT_PORT, MQTT_USER, MQTT_PASS, SUBSYSTEM_PATH
from DecenTT.MQTT.logger import logger, logging
from tqdm import tqdm


class Client:

    """
        Should not inherit the mqtt.Client.
        should provide the wrapped class for mqtt client with logging
        should write logs
    """

    def __init__(
        self,
        host: str = MQTT_HOST,
        port: int = int(MQTT_PORT),
        username: str = MQTT_USER,
        password: str = MQTT_PASS,
        log_path: str = f"{SUBSYSTEM_PATH}/mqtt.log",
        on_connect=None,
        on_message=None,
        on_publish=None,
            **kwargs
    ) -> None:
        self.__log_path = log_path
        client = mqtt.Client()
        client.username_pw_set(username=username, password=password)

        if on_connect == None:
            client.on_connect = self.__on_connect
        else:
            client.on_connect = on_connect

        if on_message == None:
            client.on_message = self.__on_message
        else:
            client.on_message = on_message

        if on_publish == None:
            client.on_publish = self.__on_publish
        else:
            client.on_publish = on_publish

        client.connect(host=host, port=port)
        logger.info(f"Connected to {client}!")
        self.client = client

    def __on_connect(self, client, userdata, flags, rc):
        logger.info(f"connecting to {client}, userdata = {userdata}, flags = {flags}")
        for i in tqdm(range(100)):
            pass
        logger.info(f"Connected to {client}!")

    def __on_message(self, client, userdata, msg):
        logger.debug(f"{msg.topic} got :\t {(msg.payload)}")

    def __on_publish(self, client, userdata, result):
        logger.debug(f"Published! result=\t{result}")

    def get_client(self) -> mqtt.Client:
        return self.client

    def show_logs(self):
        self.client.subscribe("$SYS/#", 1)
        try:
            self.client.loop_forever()
        except KeyboardInterrupt:
            logger.warning(f"Interrupted with KeyBoard")
            self.client.unsubscribe("$SYS/#")

    def write_logs(self, path: str = None, level: str = "DEBUG"):
        if(path == None):
            path = self.__log_path
        file_handler = logging.FileHandler(path)
        levels = {
            "DEBUG": logging.DEBUG,
            "INFO": logging.INFO,
            "ERROR": logging.ERROR,
            "WARNING": logging.WARNING,
            "CRITICAL": logging.CRITICAL,
        }
        file_handler.setLevel(level=levels[level])
        logger.addHandler(file_handler)
        self.show_logs()
        logger.removeHandler(file_handler)



from DecenTT.DecenTTErrors import InvalidCallback, TopicIncompatibleType
from DecenTT.IPFS import Client as iClient  # Equipped with list of topics
from DecenTT.MQTT import Client as mClient


class Client:
    """
        Which Client ??
        0:          MQTT Client
        1:          IPFS Client

    """

    def __init__(
            self,
            host: str,
            port: int = 1883,
            username: str = None,
            password: str = None,
            on_connect=None,
            on_publish=None,
            on_message=None,
            log_path: str = ".",
            keyword: str = "secure/",
            **kwargs
    ) -> None:
        self.__keyword = keyword
        self.clients = [
            mClient(
                host=host,
                port=port,
                username=username,
                password=password,
                on_connect=on_connect,
                on_publish=on_publish,
                on_message=on_message,
                log_path=log_path,
                **kwargs
            ),

            iClient(**kwargs)
        ]

        self.iClient_subscriptions = None


    # Mock Switcher
    def __which(self, topic: str) -> object:
        if self.__keyword in topic: return self.clients[1]
        return self.clients[0]

    def __map_clients_to_topic(self, topic: list) -> tuple:
        client_0_list = list()
        client_1_list = list()
        for t_i in topic:
            if self.__keyword in str(t_i[0]):
                client_1_list.append(t_i[0])
            else:
                client_0_list.append(t_i)

        return client_0_list, client_1_list

    def subscribe(self, topic: any, callback, **kwargs):

        # check for the validity of the topic as data_structure
        if not (isinstance(topic, list) or isinstance(topic, str)):
            raise TopicIncompatibleType

        # check for valid callback
        if not ('function' in str(type(callback))):
            raise InvalidCallback(callback)

        if isinstance(topic, list):
            c0_topics, c1_topics = self.__map_clients_to_topic(topic= list(topic))
            self.clients[0].subscribe(topic=c0_topics, callback=callback)
            self.iClient_subscriptions = self.clients[1].subscribe(topic=c1_topics, callback=callback)
            return

        client = self.__which(topic=topic)
        if(isinstance(client, iClient)):
            self.iClient_subscriptions = client.subscribe(topic=topic, callback=callback, **kwargs)
        client.subscribe(topic=topic, callback=callback, **kwargs)

    def __kill_all_threads(self):
        thread_list = [sub.thread for sub in self.iClient_subscriptions]
        for thread in thread_list:
            #print(f"Closing thread: {thread}")
            thread.terminate()
            #print(f"Terminating thread: {thread}")
            thread.join()
            #print(f"Closed thread: {thread}")

    def stop(self):
        self.__kill_all_threads()

    def publish(self, topic: str, payload: str, **kwargs):
        client = self.__which(topic=topic)
        client.publish(topic=topic, payload=payload, **kwargs)

    def loop(self, **kwargs):
        self.clients[0].client.loop(**kwargs)

    def loop_start(self):
        self.clients[0].client.loop_start()

    def loop_stop(self, **kwargs):
        self.clients[0].client.loop_stop(**kwargs)
        self.__kill_all_threads()

    def loop_forever(self, **kwargs):
        self.clients[0].client.loop_forever(**kwargs)



import os
from DecenTT.env.load_env import load_env


# Load the dot env file

# root_path = os.path.dirname(os.path.abspath(__file__))
# load_env(path=root_path + "/../.env")
from dotenv import load_dotenv

load_dotenv()


# Declare all the exportable variables here
MQTT_HOST = str(os.environ.get("MQTT_HOST"))
MQTT_PORT = str(os.environ.get("MQTT_PORT"))
MQTT_USER = str(os.environ.get("MQTT_USER"))
MQTT_PASS = str(os.environ.get("MQTT_PASS"))
SUBSYSTEM_PATH = str(os.environ.get("SUBSYSTEM_PATH"))

LOG_PATH = str(os.environ.get("LOG_PATH"))

DEFAULT_HOST = str(os.environ.get("DEFAULT_HOST"))




import platform


class TopicIncompatibleType(Exception):
    """
        Subjected Topic is not a compatible type
    """
    def __str__(self):
        return f"Topic should either be a string or a list!"


class UnsupportedPlatform(Exception):
    """
        Subjected Platform is currently not supported by the IPFS installer.
    """

    def __str__(self):
        return f"{platform.system()} is currently not supported by the IPFS installer."


class InvalidCallback(Exception):
    """
        Subjected Callback is not a valid callback
    """

    def __init__(self, callback) -> None:
        self.called_callback = callback

    def __str__(self):
        return f"{self.called_callback.__name__} is not a valid Function!"



import os
from DecenTT.IPFS.locales.IPFSLocalesExceptions import *


def onLoadEnv(path: str = None):
    import dotenv
    dotenv.load_dotenv() if path is None else dotenv.load_dotenv(dotenv_path=path)


def getBootstrapNodes(path: str = None) -> list:
    import re
    onLoadEnv(path=path)

    pat = re.compile("BOOTSTRAP")
    bs_keys = list(filter(pat.match, list(os.environ)))

    nodes = list()
    try:
        for i, key in enumerate(bs_keys):
            nodes.append(os.environ[key])
    except Exception as e:
        raise BootstrapNodeError(error=e)

    return nodes


def getIPFSHost() -> str:
    try:
        onLoadEnv()
        return os.environ["DEFAULT_HOST"]
    except Exception as e:
        return "127.0.0.1:5001"



import setuptools
import os

def get_deps()-> list:
    thelibFolder = os.path.dirname(os.path.realpath(__file__))
    requirementPath = thelibFolder + '/requirements.txt'
    install_requires = [] # Here we'll get: ["gunicorn", "docutils>=0.3", "lxml==0.5a7"]
    if os.path.isfile(requirementPath):
        with open(requirementPath) as f:
            install_requires = f.read().splitlines()

    return install_requires


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="DecenTT",
    version="0.0.5",
    author="Amey Mahadik",
    author_email="ameyarm@gmail.com",
    description="Decentralized Telemetry Transport (DecenTT)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/saapo-ka-baadshah/DecenTT",
    project_urls={
        "Bug Tracker": "https://github.com/saapo-ka-baadshah/DecenTT/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
)



import subprocess
from DecenTT.IPFS.locales.bootstrap import BStrapper
from DecenTT.IPFS.pubsub import Client
import requests


class Daemon:
    def __init__(self, **kwargs):
        self._kwargs = kwargs
        self.daemon = self.__start_daemon(**self._kwargs)

    @staticmethod
    def __start_daemon(
            pubsub_mode=True,
            **kwargs
    ):
        if not pubsub_mode:
            return subprocess.Popen(["ipfs", "daemon"], **kwargs)
        return subprocess.Popen(["ipfs", "daemon", "--enable-pubsub-experiment"], **kwargs)

    def stop(self):
        self.daemon.kill()
        _sd_process = subprocess.Popen(["ipfs", "shutdown"])

    def reload(self):
        self.stop()
        self.daemon = self.__start_daemon(**self._kwargs)


class Peer:
    def __init__(self):
        self.details = requests.post(
            f"http://127.0.0.1:5001/api/v0/id"
        ).json()
        self.id = self.details["ID"]
        self.peers = requests.post(
            f"http://127.0.0.1:5001/api/v0/swarm/peers"
        ).json()
        self.bootstraps = requests.post(
            f"http://127.0.0.1:5001/api/v0/bootstrap/list"
        ).json()
        self.bootstrap_peers = set([self.__get_peerId(address=str(x)) for x in self.bootstraps["Peers"]])

    def __get_peerId(self, address: str):
        domain_list = address.split("/")
        return domain_list[-1]



import json
import ipfsApi

import requests

from DecenTT.IPFS.IPFSExceptions import *
from DecenTT.IPFS.locales import getIPFSHost
from DecenTT.IPFS.config import ENDPOINTS
from requests import post
from urllib.parse import urlencode
from DecenTT.IPFS.threader import ThreadDecorator, ProcessDecorator
from DecenTT.IPFS.standard import IPFSResponse

from DecenTT.env.config import DEFAULT_HOST


class Client:
    """
    >>> c = Client()
    >>> c.publish("abc", "Hello Worlds!!!")
    <Response [200]>
    """

    # Subscription Class:
    class Subscription:
        def __init__(self, topic: str, callback, client, host: str = DEFAULT_HOST, **kwargs):
            self.host = host
            self.topic = topic
            self.client = client
            if (not 'function' in str(type(callback))) or callback is None:
                raise InvalidCallback(callback=callback)
            else:
                self.__callback = callback
                self.thread = self.__run(topic=self.topic, callback=self.__callback)

        #@ThreadDecorator.thread
        @ProcessDecorator.thread
        def __callback_thread(self, resp: requests.Response, callback) -> None:
            for line in resp.iter_lines(decode_unicode=True):
                payload = json.loads(line)

                """
                    Match the standards of MQTT here
                        arg1: client : ipfsApi.Client
                        arg2: userdata : from JSON response of IPFS HTTP API
                        arg3: msg : from JSON 
                """
                ipfs_resp = IPFSResponse(in_dict=payload)
                __client = self.client
                __usrdata = ipfs_resp.userdata
                __msg = ipfs_resp.msg

                callback(__client, __usrdata, __msg)

        def __run(self, topic: str, callback):
            """
                Refer to IPFS HTTP API Docs: https://docs.ipfs.io/reference/http/api/#api-v0-pubsub-sub
            """

            ep = ENDPOINTS["pubsub"]["sub"]
            args = [
                ("arg", topic),
                # ("discover", None)                # Deprecated
            ]
            url = f"http://{self.host}{ep}?{urlencode(args)}"

            # Create Request to: IPFS HTTP API
            resp = post(url=url, stream=True)
            # Check for Encoding
            if resp.encoding is None:
                resp.encoding = 'utf-8'

            return self.__callback_thread(resp, callback)

        def close(self):
            del self

    def __init__(self) -> None:
        self.host = getIPFSHost()
        self.host_address = self.host.split(":")[0]
        self.host_port = int(self.host.split(":")[1])
        self.ipfs_client = ipfsApi.Client(self.host_address, self.host_port)

    def subscribe(self, topic: any, callback, **kwargs):
        if not (isinstance(topic, list) or isinstance(topic, str)):
            raise TopicIncompatibleType
        if isinstance(topic, list):
            return [self.Subscription(topic=t_i, callback=callback, client=self) for t_i in list(topic)]
        return self.Subscription(topic=topic, callback=callback, client=self)

    def publish(self, topic: str, payload: str, **kwargs) -> None:
        """
            Refer to IPFS HTTP API Docs: https://docs.ipfs.io/reference/http/api/#api-v0-pubsub-pub
        """
        ep = ENDPOINTS["pubsub"]["pub"]

        args = [
            ("arg", topic),
            ("arg", payload)
        ]

        url = f"http://{self.host}{ep}?{urlencode(args)}"

        resp = post(url=url)

        if resp.status_code == 200 and resp.reason == "OK":
            return resp

        resp.raise_for_status()



from paho.mqtt import client

from DecenTT.MQTT.MQTTErrors import InvalidCallback
from DecenTT.MQTT.logger import logger


class Client:
    def __init__(
            self,
            host: str,
            port: int = 1883,
            username: str = None,
            password: str = None,
            on_connect=None,
            on_publish=None,
            on_message=None,
            log_path: str = ".",
            **kwargs
    ) -> None:
        self.client = client.Client()

        self.__host = host
        self.__port = port
        self.__uname = username
        self.__pass = password
        self.__on_connect = on_connect if not on_connect is None else None
        self.__on_message = on_message if not on_message is None else self.__int_on_message
        self.__on_publish = on_publish if not on_publish is None else None

        if not self.__on_connect is None:
            self.client.on_connect = self.__on_connect

        if not self.__on_publish is None:
            self.client.on_publish = self.__on_publish

        self.client.on_message = self.__on_message

        # Set username password
        if self.__uname:
            self.client.username_pw_set(username=self.__uname, password=self.__pass)

        # Connect to the Client
        self.client.connect(self.__host, self.__port)

    def __int_on_message(self, client, userdata, msg):
        logger.info(f"{msg.topic} <- {msg.payload}")

    def subscribe(self, topic, callback, **kwargs):
        # check for valid callback
        if not ('function' in str(type(callback))):
            raise InvalidCallback(callback)
        self.client.on_message = callback
        self.client.subscribe(topic=topic, **kwargs)

    def publish(self, topic: str, payload: str, **kwargs):
        self.client.publish(topic=topic, payload=payload, **kwargs)



from DecenTT.MQTT.client import Client
from DecenTT.MQTT.logger import logging
from DecenTT.MQTT.logger import logger
from DecenTT.MQTT.MQTTErrors import MQTTValueError
from DecenTT.env.config import SUBSYSTEM_PATH
import threading
import time
import ctypes
import sys

class Publisher:

    def __init__(self, client: Client, topic: str, qos :int = 1,
    log_path:str=f"{SUBSYSTEM_PATH}/runnable.log",
    log_level:str="DEBUG") -> None:
        self.client = client.client
        self.topic = topic
        self.__qos = qos
        self.__pub_val = None
        self.__temp = None
        self.close_runnable = False
        self.thread = None
        self.__log_path = log_path
        self.__log_level = log_level

    def publish(self, payload:any):
        self.client.publish(
            topic = self.topic,
            payload=payload,
            qos= self.__qos
            )


    @property
    def publisherVariable(self):
        return self.__pub_val

    @publisherVariable.setter
    def publisherVariable(self, val):
        if val != self.__pub_val:
            self.client.publish(self.topic, val, qos= self.__qos)
            self.__pub_val = val

    def __thread(self, interval:int):

        while True:

            val_list = ctypes.cast(id(self.__temp), ctypes.py_object).value

            if(len(val_list) == 0):
                raise MQTTValueError
            val = ", ".join(map(str, val_list))

            self.client.publish(self.topic, val, qos= self.__qos)
            logger.debug(f"{val} published")
            if(self.close_runnable): 
                logger.warning("Runnable Closed")
                logger.removeHandler(self.__file_handler)
                sys.exit()

            time.sleep(interval)


    def runnable_publish(self, val:list, interval:int):
        """
            Pass the variable as a non-mutable object
        """
        self.__temp = val


        self.__file_handler = logging.FileHandler(self.__log_path)
        levels = {
            "DEBUG": logging.DEBUG,
            "INFO": logging.INFO,
            "ERROR": logging.ERROR,
            "WARNING": logging.WARNING,
            "CRITICAL": logging.CRITICAL,
        }
        self.__file_handler.setLevel(level=levels[self.__log_level])
        logger.addHandler(self.__file_handler)


        logger.debug(f"runnable publish called")
        self.thread = threading.Thread(target=self.__thread, args=(interval, ))
        logger.debug(f"{self.thread} created")
        self.thread.start()
